/*
	Vertris v0.3.2
	Copyright (C) 2006 - 2007, Andrs Rassol
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <string.h>

#include <allegro.h>

#include "vertris.h"

extern SAMPLE Move;

extern int Redraw_All;

extern FXVertris[VERTRIS_W][VERTRIS_H];

char Vertris[VERTRIS_W][VERTRIS_H];
Vertromino Vertro;

char Max_Name[MAX_HISCORES][NAME_LENGTH];
unsigned int Lines, Max_Lines[MAX_HISCORES], Score, Max_Score[MAX_HISCORES], Level;
char Fullscreen, Pause, Sound;

static char End_Game, Sound_Init;

static volatile unsigned int Ticks, Key_Up;

char Ghosting_FX;

static void add_tick() {Ticks++;} END_OF_STATIC_FUNCTION(add_tick);

void reset_ticks() {Ticks = 0;} END_OF_FUNCTION(reset_ticks);

static void keypress_watcher(int scancode)
{
	if (scancode & 0x80)
		Key_Up = 1;
}
END_OF_FUNCTION(keypress_watcher)

static void reset_key_up() {Key_Up = 0;} END_OF_FUNCTION(reset_key_up)

static void init()
{
	srand(time(NULL));

	set_uformat(U_ASCII);

	if (allegro_init() != 0)
		exit(1)	;

	load_prefs();

	init_output();

	install_keyboard();
	install_timer();

	set_keyboard_rate(H_SPEED, H_SPEED);

	set_window_title("Vertris");

	LOCK_FUNCTION(keypress_watcher);
	LOCK_FUNCTION(reset_key_up);
	LOCK_VARIABLE(Key_Up);
	keyboard_lowlevel_callback = keypress_watcher;

	LOCK_FUNCTION(add_tick);
	LOCK_FUNCTION(reset_ticks);
	LOCK_VARIABLE(Ticks);
	install_int(add_tick, V_SPEED);

	gen_vertro();
}

void new_game()
{
	memset(Vertris, 0, VERTRIS_W * VERTRIS_H);
	memset(FXVertris, 0, VERTRIS_W * VERTRIS_H);

	Lines = 0;
	Score = 0;
	Ticks = 0;
	Level = 0;
	Redraw_All = 1;

	gen_vertro();
	gen_vertro();
}

static void input()
{	
	if (key[KEY_P] || key[KEY_PAUSE])
	{
		Pause = !Pause;
		Redraw_All = 1;
		reset_ticks();
	}

	if (key[KEY_ESC] || key[KEY_Q])
		End_Game = 1;

	if (key[KEY_M])
	{
		Sound = !Sound;
		
		if (!Sound_Init)
		{
			if (install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL) != 0)
				Sound = 0;
			else
				Sound = 1;

			Sound_Init = 1;
		}
	}
		
	if (key[KEY_ALT] && key[KEY_ENTER])
	{
		Pause = 1;
		if (Fullscreen)
			set_gfx_mode(GFX_AUTODETECT_WINDOWED, WIDTH, HEIGHT, 0, 0) == 0;
		else
			set_gfx_mode(GFX_AUTODETECT_FULLSCREEN, WIDTH, HEIGHT, 0, 0) == 0;

		Fullscreen = !Fullscreen;
		Redraw_All = 1;
	}	

	if (!Pause)
	{
		if (key[KEY_R])
		{
			if (is_hiscore())
				hiscores();
			else
				new_game();
		}

		if ( (key[KEY_DOWN] || key[KEY_S] || key[KEY_5_PAD]) && valid_pos(Vertro.x, Vertro.y - 1))
		{
			Vertro.y--;
			output();
		}

		if ( (key[KEY_RIGHT] || key[KEY_D] || key[KEY_6_PAD]) && valid_pos(Vertro.x + 1, Vertro.y))
		{
			Vertro.x++;
			output();
			play_sound(Move);
		}

		if ( (key[KEY_LEFT] || key[KEY_A] || key[KEY_4_PAD]) && valid_pos(Vertro.x - 1, Vertro.y) )
		{
			Vertro.x--;
			output();
			play_sound(Move);
		}

		if (key[KEY_UP] || key[KEY_W] || key[KEY_8_PAD])
		{
			vertro_rotate(0);
			play_sound(Move);
		}
		else if (key[KEY_0_PAD])
		{
			vertro_rotate(1);
			play_sound(Move);
		}

		if ( (key[KEY_SPACE] || key[KEY_PLUS_PAD]) && Key_Up )
		{
			while (valid_pos(Vertro.x, Vertro.y - 1))
			{
				Ghosting_FX = Ghosting_FX > 5 ? Ghosting_FX : Ghosting_FX + 1;
				Vertro.y--;
				output();
				rest(10);
			}

			output();

			while (Ghosting_FX > 0 && --Ghosting_FX)
				output();

			vert_setpiece();
			reset_ticks();
		}
	}

	reset_key_up();

	clear_keybuf();
}

static void process()
{
	int d_lines, speed;

	if (Pause)
		reset_ticks();

	while (!valid_pos(Vertro.x, Vertro.y - Ticks) && Ticks > 1)
		Ticks--;

	d_lines = Lines;

	if (valid_pos(Vertro.x, Vertro.y - Ticks))
		Vertro.y -= Ticks;	
	else
	{
		play_sound(Move);
		vert_setpiece();
		gen_vertro();
		vert_lineclear();
		Level = level();
	}

	speed = V_SPEED - Lines - Level * 10;

	if (speed >= 100)
		install_int(add_tick, speed);
	else
		install_int(add_tick, 100);

	d_lines = Lines - d_lines;

	switch (d_lines)
	{
		case 0:	break;
		case 1:	Score += 40; break;
		case 2:	Score += 100; break;
		case 3:	Score += 300; break;
		case 4:	Score += 1200;
	}

	reset_ticks();
}

int main()
{
	init();
	new_game();

	load_state();

	while (!End_Game)
	{
		while (!keypressed())
		{
			output();
			process();
			rest(10);
		}

		input();
	}

	save_state();

	return 0;
}
END_OF_MAIN()
